from __future__ import annotations
from random import choice
from typing import TYPE_CHECKING, Callable, Literal, Sequence

from negmas.common import PreferencesChange
from negmas.gb.components import GBComponent
from negmas.preferences import (
    BaseUtilityFunction,
    InverseUFun,
    PresortingInverseUtilityFunction,
    RankOnlyUtilityFunction,
)

__all__ = ["UtilityInverter", "UtilityBasedOutcomeSetRecommender"]

if TYPE_CHECKING:
    from negmas.gb import GBNegotiator, GBState
    from negmas.outcomes import Outcome

    from .selectors import OfferSelectorProtocol


def make_inverter(
    ufun: BaseUtilityFunction,
    ufun_inverter: Callable[[BaseUtilityFunction], InverseUFun] | None = None,
    rank_only: bool = False,
    max_cardinality: int | float = float("inf"),
) -> InverseUFun:
    """
    Creates an `InverseUFun` object from the given ufun with appropriate type if the type is not given

    Args:
        ufun (BaseUtilityFunction): The ufun to invert
        rank_only (bool): If True, only the relative ranks of outcomes will be used in inversion not the values themselves.
        ufun_inverter (Callable[[BaseUtilityFunction], InverseUFun] | None): An optional factory to generate a `InverseUFun` from a `BaseUtilityFunction` .
        max_cardinality (int): The maximum cardinality at which we switch to using a `SamplingInverseUtilityFunction`
    """
    if rank_only:
        ufun = RankOnlyUtilityFunction(ufun, randomize_equal=False, name=ufun.name)
    if ufun_inverter:
        return ufun_inverter(ufun)
    return PresortingInverseUtilityFunction(ufun, rational_only=True)
    # return (
    #     SamplingInverseUtilityFunction(ufun)
    #     if ufun.outcome_space is None
    #     or ufun.outcome_space.is_discrete() and ufun.outcome_space.cardinality >= max_cardinality
    #     else PresortingInverseUtilityFunction(ufun)
    # )


class UtilityBasedOutcomeSetRecommender(GBComponent):
    """
    Recommends a set of outcome appropriate for proposal

    """

    def __init__(
        self,
        rank_only: bool = False,
        ufun_inverter: Callable[[BaseUtilityFunction], InverseUFun] | None = None,
        max_cardinality: int | float = float("inf"),
        eps: float = 0.0001,
        inversion_method: Literal["min"]
        | Literal["max"]
        | Literal["one"]
        | Literal["some"]
        | Literal["all"] = "some",
    ):
        super().__init__()
        self._rank_only = rank_only
        self._max_cardinality = max_cardinality
        self._ufun_inverter = ufun_inverter
        self._inv_method = None
        self._inversion_method = inversion_method
        self.eps = eps
        self.inv: InverseUFun | None = None
        self.min = self.max = self.best = None
        self._inv_method = None

    def set_negotiator(self, negotiator: GBNegotiator) -> None:
        super().set_negotiator(negotiator)
        self.inv: InverseUFun | None = None
        self.min = self.max = self.best = None
        self._inv_method = None

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if self._negotiator is None:
            raise ValueError("Unknown negotiator in a component")
        ufun = self._negotiator.ufun

        if not ufun:
            self.inv = None
            self.min = self.max = self.best = None
            return
        self.inv = make_inverter(
            ufun, self._ufun_inverter, self._rank_only, self._max_cardinality
        )
        if self._inversion_method == "one":
            self._inv_method = self.inv.one_in
            self._single_inv_return = True
        elif self._inversion_method == "min":
            self._inv_method = self.inv.worst_in
            self._single_inv_return = True
        elif self._inversion_method == "max":
            self._inv_method = self.inv.best_in
            self._single_inv_return = True
        elif self._inversion_method == "some":
            self._inv_method = self.inv.some
            self._single_inv_return = False
        elif self._inversion_method == "all":
            self._inv_method = self.inv.all  # type: ignore
            self._single_inv_return = False
        else:
            raise ValueError(f"Unknown selectortype: {self._inversion_method}")

        _worst, self.best = ufun.extreme_outcomes()
        self.min, self.max = float(ufun(_worst)), float(ufun(self.best))
        if self.min < ufun.reserved_value:
            self.min = ufun.reserved_value

    def before_proposing(self, state: GBState):
        if self._negotiator is None:
            raise ValueError("Unknown negotiator in a component")
        if self.inv is None or self.max is None or self.min is None:
            # warn(
            #     f"It seems that on_prefrences_changed() was not called until propose for a ufun of type {self._negotiator.ufun.__class__.__name__}"
            #     f" for negotiator {self._negotiator.id} of type {self.__class__.__name__}",
            #     NegmasUnexpectedValueWarning,
            # )
            self.on_preferences_changed([PreferencesChange()])
        if self.inv is None or self.max is None or self.min is None:
            raise ValueError(
                "Failed to find an invertor, a selector, or exreme outputs"
            )
        if not self.inv.initialized:
            self.inv.init()

    def scale_utilities(self, urange: tuple[float, ...]) -> tuple[float, ...]:
        """
        Scales given utilities to the range of the ufun.

        Remarks:

            - Assumes that the input utilities are in the range [0-1] no matter what
              is the range of the ufun.
            - Subtracts the `tolerance` from the first and adds it to the last utility value
              which slightly enlarges the range to account for small rounding errors
        """
        if self._negotiator is None:
            raise ValueError("Unknown negotiator in a component")
        if self.max is None or self.min is None or self.best is None:
            # warn(
            #     f"It seems that on_prefrences_changed() was not called until propose for a ufun of type {self._negotiator.ufun.__class__.__name__}"
            #     f" for negotiator {self._negotiator.id} of type {self.__class__.__name__}",
            #     NegmasUnexpectedValueWarning,
            # )
            self.on_preferences_changed([PreferencesChange()])
        if self.max is None or self.min is None:
            raise ValueError("Cannot find extreme outcomes.")
        adjusted = [(self.max - self.min) * _ + self.min for _ in urange]
        if adjusted:
            adjusted[0] -= self.eps
            adjusted[-1] += self.eps
        return tuple(adjusted)

    @property
    def tolerance(self):
        return self.eps

    @property
    def ufun_max(self):
        return self.max

    @property
    def ufun_min(self):
        return self.min

    def __call__(
        self, urange: tuple[float, float], state: GBState
    ) -> Sequence[Outcome]:
        """
        Receives a normalized [0-> 1] utility range and returns a utility range relative to the ufun taking the tolerance _eps into account

        Remarks:

            - This method calls `scale_utilities` on the input range
        """
        if self._negotiator is None:
            raise ValueError("Unknown negotiator in a component")
        urange = self.scale_utilities(urange)
        outcomes = self._inv_method(urange, normalized=False)  # type: ignore
        if outcomes is None:
            return []
        if self._single_inv_return:
            return [outcomes]  # type: ignore
        return outcomes


class UtilityInverter(GBComponent):
    """
    A component that can recommend an outcome based on utility
    """

    def set_negotiator(self, negotiator: GBNegotiator) -> None:
        super().set_negotiator(negotiator)
        self.recommender.set_negotiator(negotiator)

    def __init__(
        self,
        *args,
        offer_selector: OfferSelectorProtocol
        | Literal["min"]
        | Literal["max"]
        | None = None,
        **kwargs,
    ):
        if offer_selector is None:
            type_ = "one"
        elif isinstance(offer_selector, Callable):
            type_ = "some"
        else:
            type_ = offer_selector
        self.recommender = UtilityBasedOutcomeSetRecommender(
            *args, inversion_method=type_, **kwargs
        )
        self.selector: Callable[[Sequence[Outcome], GBState], Outcome | None] | None
        self.selector = (
            None if not isinstance(offer_selector, Callable) else offer_selector
        )
        self.set_negotiator(None)  # type: ignore (It is OK. We do not really need to pass this at all here.)

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        self.recommender.on_preferences_changed(changes)

    def before_proposing(self, state: GBState):
        self.recommender.before_proposing(state)

    def scale_utilities(self, urange):
        return self.recommender.scale_utilities(urange)

    @property
    def tolerance(self):
        return self.recommender.eps

    @property
    def ufun_max(self):
        return self.recommender.max

    @property
    def ufun_min(self):
        return self.recommender.min

    def __call__(self, urange: tuple[float, float], state: GBState) -> Outcome | None:
        """
        Receives a normalized [0-> 1] utility range and returns a utility range relative to the ufun taking the tolerance _eps into account

        Remarks:

            - This method calls `scale_utilities` on the input range
        """
        outcomes = self.recommender(urange, state)
        if not outcomes:
            return None
        if len(outcomes) == 1:
            return outcomes[0]
        if self.selector is None:
            return choice(outcomes)
        outcome = self.selector(outcomes, state)
        if not outcome:
            return self.recommender.best
        return outcome



from __future__ import annotations
from typing import TYPE_CHECKING, Any

from .entity import Entity

if TYPE_CHECKING:
    from .world import World


__all__ = ["StatsMonitor", "WorldMonitor"]


class StatsMonitor(Entity):
    """A monitor object capable of receiving stats of a world"""

    def init(self, stats: dict[str, Any], world_name: str):
        """Called to initialize the monitor before running first step"""

    def step(self, stats: dict[str, Any], world_name: str):
        """Called at the END of every simulation step"""


class WorldMonitor(Entity):
    """A monitor object capable of monitoring a world. It has read/write access to the world"""

    def init(self, world: World):
        """Called to initialize the monitor before running first step"""

    def step(self, world: World):
        """Called at the END of every simulation step"""



from __future__ import annotations
from abc import ABC, abstractmethod
from collections import defaultdict
from random import randint, random, sample, shuffle
from typing import Any, Callable, Collection

import numpy as np

from negmas import MechanismState, Negotiator, NegotiatorMechanismInterface
from negmas.gb.negotiators.base import GBNegotiator
from negmas.outcomes import Issue, dict2outcome, make_issue
from negmas.preferences import LinearUtilityFunction, UtilityFunction
from negmas.sao.negotiators import RandomNegotiator, SAONegotiator
from negmas.serialization import to_flat_dict
from negmas.situated import (
    Action,
    Agent,
    AgentWorldInterface,
    Breach,
    Contract,
    RenegotiationRequest,
    World,
)


# just repeating the code from the previous tutorial
class AWI(AgentWorldInterface):
    @property
    def n_negs(self) -> int:
        """Number of negotiations an agent can start in a step (holiday season)"""
        return self._world.neg_quota_step

    @property
    def agents(self):
        """List of all other agent IDs"""
        return list(_ for _ in self._world.agents.keys() if _ != self.agent.id)

    def request_negotiation(
        self, partners: list[str], negotiator: SAONegotiator | GBNegotiator
    ) -> bool:
        """A convenient way to request negotiations"""
        _world: TripsWorld = self._world  # type: ignore
        if self.agent.id not in partners:
            partners.append(self.agent.id)
        req_id = self.agent.create_negotiation_request(
            issues=tuple(_world.ISSUES),
            partners=partners,
            negotiator=negotiator,
            annotation=dict(),
            extra=dict(negotiator_id=negotiator.id),
        )
        return self.request_negotiation_about(
            issues=tuple(_world.ISSUES), partners=partners, req_id=req_id
        )


class Person(Agent[AWI], ABC):
    @abstractmethod
    def step(self):
        ...

    @abstractmethod
    def init(self):
        ...

    @abstractmethod
    def respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        mechanism: NegotiatorMechanismInterface,
    ) -> Negotiator | None:
        ...

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        issues: list[Issue],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        role: str | None,
        req_id: str | None,
    ) -> Negotiator | None:
        return self.respond_to_negotiation_request(initiator, partners, mechanism)

    def on_neg_request_rejected(self, req_id: str, by: list[str] | None):
        pass

    def on_neg_request_accepted(
        self, req_id: str, mechanism: NegotiatorMechanismInterface
    ):
        pass

    def on_negotiation_failure(
        self,
        partners: list[str],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        state: MechanismState,
    ) -> None:
        pass

    def on_negotiation_success(
        self, contract: Contract, mechanism: NegotiatorMechanismInterface
    ) -> None:
        pass

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: list[Breach]
    ) -> RenegotiationRequest | None:
        pass

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: list[Breach], agenda: RenegotiationRequest
    ) -> Negotiator | None:
        pass

    def on_contract_executed(self, contract: Contract) -> None:
        pass

    def on_contract_breached(
        self, contract: Contract, breaches: list[Breach], resolution: Contract | None
    ) -> None:
        pass


class TripsWorld(World[AWI, Person]):
    ISSUES = (
        make_issue((0.0, 1.0), "cost"),
        make_issue(2, "active"),
        make_issue((1, 7), "duration"),
    )

    def __init__(self, *args, **kwargs):
        """Initialize the world"""
        kwargs["awi_type"] = AWI
        kwargs["negotiation_quota_per_step"] = kwargs.get(
            "negotiation_quota_per_step", 8
        )
        kwargs["force_signing"] = True
        kwargs["default_signing_delay"] = 0
        super().__init__(*args, **kwargs)
        self._contracts: dict[int, list[Contract]] = defaultdict(list)
        self._total_utility: dict[str, float] = defaultdict(float)
        self._ufuns: dict[str, UtilityFunction] = dict()
        self._breach_prob: dict[str, float] = dict()

    def join(
        self, x, simulation_priority: int = 0, ufun=None, breach_prob=None, **kwargs
    ):
        """Define the ufun and breach-probability for each agent"""
        super().join(x, **kwargs)
        weights = (np.random.rand(len(self.ISSUES)) - 0.5).tolist()
        x.ufun = (
            LinearUtilityFunction(
                weights=weights, reserved_value=0.0, issues=self.ISSUES
            )
            if ufun is None
            else ufun
        )
        self._ufuns[x.id] = x.ufun
        self._breach_prob[x.id] = random() * 0.1 if breach_prob is None else breach_prob

    def simulation_step(self, stage: int = 0):
        """What happens in this world? Nothing"""
        pass

    def get_private_state(self, agent: Agent) -> dict:
        """What is the information available to agents? total utility points"""
        return dict(total_utility=self._total_utility[agent.id])

    def execute_action(
        self, action: Action, agent, callback: Callable | None = None
    ) -> bool:
        """Executing actions by agents? No actions available"""
        ...

    def on_contract_signed(self, contract: Contract) -> bool:
        """Save the contract to be executed in the following hoiday season (step)"""
        if not super().on_contract_signed(contract):
            return False
        self._contracts[self.current_step + 1].append(contract)
        return True

    def executable_contracts(self) -> Collection[Contract]:
        """What contracts are to be executed in the current step?
        Ones that were signed the previous step"""
        return self._contracts[self.current_step]

    def order_contracts_for_execution(
        self, contracts: Collection[Contract]
    ) -> Collection[Contract]:
        """What should be the order of contract execution? Random"""
        contracts = list(contracts)
        shuffle(contracts)
        return contracts

    def start_contract_execution(self, contract: Contract) -> set[Breach] | None:
        """What should happen when a contract comes due?
        1. Find out if it will be breached
        2. If not, add to each agent its utility from the trip
        """
        breaches = []
        for aid in contract.partners:
            if random() < self._breach_prob[aid]:
                breaches.append(
                    Breach(
                        contract,
                        aid,
                        "breach",
                        victims=[_ for _ in contract.partners if _ != aid],
                    )
                )
        if len(breaches) > 0:
            return set(breaches)
        for aid in contract.partners:
            if not isinstance(contract.agreement, dict):
                continue
            self._total_utility[aid] += self._ufuns[aid](
                dict2outcome(contract.agreement, issues=tuple(self.ISSUES))
            )
        return set()

    def complete_contract_execution(
        self, contract: Contract, breaches: list[Breach], resolution: Contract
    ) -> None:
        """What happens if a breach was resolved? Nothing. They cannot"""
        pass

    def delete_executed_contracts(self) -> None:
        """Removes all contracts for the current step"""
        if self._current_step in self._contracts.keys():
            del self._contracts[self.current_step]

    def contract_record(self, contract: Contract) -> dict[str, Any]:
        """Convert the contract into a dictionary for saving"""
        return to_flat_dict(contract)

    def breach_record(self, breach: Breach) -> dict[str, Any]:
        """Convert the breach into a dictionary for saving"""
        return to_flat_dict(breach)

    def contract_size(self, contract: Contract) -> float:
        """How good is a contract? Welfare"""
        if contract.agreement is None:
            return 0.0
        return sum(
            self._ufuns[aid](dict2outcome(contract.agreement, issues=self.ISSUES))
            for aid in contract.partners
            if isinstance(contract.agreement, dict)
        )

    def post_step_stats(self):
        for aid, agent in self.agents.items():
            self._stats[f"total_utility_{agent.name}"].append(self._total_utility[aid])


class RandomPerson(Person):
    def step(self):
        awi: AWI = self.awi  # type: ignore
        # get IDs of all ogher agents from the AWI
        agents = awi.agents
        # request the maximum number of negotiations possible
        for _ in range(awi.n_negs):
            # for each negotiation, use a random subset of partners and a random negotiator
            awi.request_negotiation(
                partners=sample(agents, k=randint(1, len(agents) - 1)),
                negotiator=RandomNegotiator(),
            )

    def init(self):
        # we need no initialization
        pass

    def respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        mechanism: NegotiatorMechanismInterface,
    ) -> Negotiator | None:
        # just us a random negotiator for everything
        return RandomNegotiator()


def test_trips_run_random():
    world = TripsWorld(n_steps=10, construct_graphs=True)
    for i in range(5):
        world.join(RandomPerson(name=f"a{i}"))
    world.run_with_progress()



"""
Common datastructures used in the outcomes module.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping

from negmas.outcomes.issue_ops import issues_from_outcomes

__all__ = [
    "Outcome",
    "PartialOutcomeDict",
    "PartialOutcomeTuple",
    "OutcomeRange",
    "check_one_and_only",
    "check_one_at_most",
    "ensure_os",
    "os_or_none",
    "DEFAULT_LEVELS",
]

if TYPE_CHECKING:
    from negmas.outcomes import OutcomeSpace

Outcome = tuple
"""An outcome is a tuple of issue values."""

PartialOutcomeDict = Mapping[int, Any]
"""A partial outcome is a simple mapping between issue INDEX and its value. Both a `tuple` and a `dict[int, Any]` satisfy this definition."""

PartialOutcomeTuple = tuple
"""A partial outcome is a partial outcome by mapping missing issues to None"""

OutcomeRange = Mapping[int, Any]
"""An outcome range is a a mapping between issue INDEX and either a value, a list of values or a Tuple with two values"""

DEFAULT_LEVELS = 10


def check_one_at_most(outcome_space, issues, outcomes) -> None:
    """Ensures that at most one of the three inputs is given (i.e. not None)"""
    if outcomes and issues:
        raise ValueError("You cannot pass `issues` and `outcomes`")
    if outcomes and outcome_space:
        raise ValueError("You cannot pass `outcome_space` and `outcomes`")
    if issues and outcome_space:
        raise ValueError("You cannot pass `issues` and `outcome_space`")


def check_one_and_only(outcome_space, issues, outcomes) -> None:
    """Ensures that one and only one of the three inputs is given (i.e. not None)"""
    if not outcomes and not issues and not outcome_space:
        raise ValueError("You must pass `outcome_spae`, `issues` or `outcomes`")
    if outcomes and issues:
        raise ValueError("You cannot pass `issues` and `outcomes`")
    if outcomes and outcome_space:
        raise ValueError("You cannot pass `outcome_space` and `outcomes`")
    if issues and outcome_space:
        raise ValueError("You cannot pass `issues` and `outcomes`")


def os_or_none(outcome_space, issues, outcomes) -> OutcomeSpace | None:
    """
    Returns an outcome space from either an outcome-space, a list of issues, a list of outcomes, or the number of outcomes

    Remakrs:
        - Precedence is in the order of paramters `outcome_space` > `issues` > `outcomes`.
        - If nothing is given, it will just return None
        - Will not copy the outcome-space if it is given
        - A `CartesianOutcomeSpace` will be created if an outcome-space is not given
        - If outcomes is given or all issues are discrete, a `DiscreteCartesianOutcomeSpace` will be created
    """
    from negmas.outcomes.outcome_space import make_os

    if outcome_space:
        return outcome_space
    if issues:
        return make_os(issues)
    if not outcomes:
        return None
    return make_os(issues_from_outcomes(outcomes))


def ensure_os(outcome_space, issues, outcomes) -> OutcomeSpace:
    """
    Returns an outcome space from either an outcome-space, a list of issues, a list of outcomes, or the number of outcomes

    Remakrs:
        - Precedence is in the order of paramters `outcome_space` > `issues` > `outcomes`.
        - If neither is given, a `ValueError` exception will be raised.
        - Will not copy the outcome-space if it is given
        - A `CartesianOutcomeSpace` will be created if an outcome-space is not given
        - If outcomes is given or all issues are discrete, a `DiscreteCartesianOutcomeSpace` will be created
    """
    from negmas.outcomes.outcome_space import make_os

    if outcome_space:
        return outcome_space
    if issues:
        return make_os(issues)
    if not outcomes:
        raise ValueError(
            "No way to create an outcome-space: outcome_space, issses, and outcomes are all None or empty"
        )
    return make_os(issues_from_outcomes(outcomes))



"""
Implements Generalized Bargaining Protocol (GB) set of mechanisms and basic negotiators.
"""

from __future__ import annotations


from .common import *
from .components import *
from .mechanisms import *
from .negotiators import *
from .controllers import *
from .evaluators import *
from .constraints import *

__all__ = (
    common.__all__
    + components.__all__
    + mechanisms.__all__
    + negotiators.__all__
    + controllers.__all__
    + evaluators.__all__
    + constraints.__all__
)



# -*- coding: utf-8 -*-
"""
Tournament generation and management.
"""

from __future__ import annotations

from .neg import *
from .tournaments import *

__all__ = tournaments.__all__ + neg.__all__ + ["neg"]
